#include"bpm_utils.h"


// Tri-Diagonal Matrix linear system solver
void TDMsolver (int n, const comple *a, comple *b, const comple *c, comple *v, comple *x)
{
        /**
         * n - number of equations
         * a - suB-diagonal (means it is the diagonal below the main diagonal) -- indexed from 1..n-1
         * b - the main diagonal
         * c - suP-diagonal (means it is the diagonal above the main diagonal) -- indexed from 0..n-2
         * v - right hand side
         * x - solution
         * NOTE: array b will be DESTROYED!!!
         */
        for (int i = 1; i < n; i++)
        {
                comple m = a[i]/b[i-1];
                b[i] = b[i] - m*c[i-1];
                v[i] = v[i] - m*v[i-1];
        }
 
        x[n-1] = v[n-1]/b[n-1];
 
        for (int i = n - 2; i >= 0; i--)
                x[i]=(v[i]-c[i]*x[i+1])/b[i];
}



void SetRHS_TM(comple *ll, comple *lc, comple *lr, int n, int b1, int b2, double dx, double lambda, double nref, double epso, double epsc) {
  /*
This function sets discretization for BPM propagation in TM polarization.
It is assumed that this operator will act on an array storing field D.
  */

  int row;
  double beta, beta0, er, el;

  beta  = 2.0*M_PI*nref/lambda;
  beta0 = 2.0*M_PI/lambda;

  // left cladding
  for(row=0;row<b1;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epso - beta*beta;
  }

  // left interface
  er  = epsc;
  el  = epso;
  row = b1;
  ll[row] = +2.0*er/(dx*dx*(el+er));
  lr[row] = +2.0*el/(dx*dx*(el+er));
  lc[row] = -2.0/(dx*dx) + beta0*beta0*(2.0*el*er)/(el+er) - beta*beta;

  // core
  for(row=b1+1;row<b2;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epsc - beta*beta;
  }

  // right interface
  er  = epso;
  el  = epsc;
  row = b2;
  ll[row] = +2.0*er/(dx*dx*(el+er));
  lr[row] = +2.0*el/(dx*dx*(el+er));
  lc[row] = -2.0/(dx*dx) + beta0*beta0*(2.0*el*er)/(el+er) - beta*beta;

  // right cladding
  for(row=b2+1;row<n;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epso - beta*beta;
  }
}

void SetRHS_TE(comple *ll, comple *lc, comple *lr, int n, int b1, int b2, double dx, double lambda, double nref, double epso, double epsc) {
  /*
This function sets discretization for BPM propagation in TE polarization.
It is assumed that this operator will act on an array storing field E.
  */

  int row;
  double beta, beta0, er, el;

  beta  = 2.0*M_PI*nref/lambda;
  beta0 = 2.0*M_PI/lambda;

  // left cladding
  for(row=0;row<b1;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epso - beta*beta;
  }

  // left interface
  er  = epsc;
  el  = epso;
  row = b1;
  ll[row] = +1.0/(dx*dx);
  lr[row] = +1.0/(dx*dx);
  lc[row] = -2.0/(dx*dx) + beta0*beta0*(el+er)/2.0 - beta*beta;

  // core
  for(row=b1+1;row<b2;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epsc - beta*beta;
  }

  // right interface
  er  = epso;
  el  = epsc;
  row = b2;
  ll[row] = +1.0/(dx*dx);
  lr[row] = +1.0/(dx*dx);
  lc[row] = -2.0/(dx*dx) + beta0*beta0*(el+er)/2.0 - beta*beta;

  // right cladding
  for(row=b2+1;row<n;row++) {
    ll[row] = +1.0/(dx*dx);
    lr[row] = +1.0/(dx*dx);
    lc[row] = -2.0/(dx*dx) + beta0*beta0*epso - beta*beta;
  }
}




void Save(const char *fname, int n, double *cx, comple *a) {
  ofstream f;
  f.open(fname);
  for(int x=0;x<n;x++) f << cx[x] <<" "<< norm(a[x]) <<" "<< real(a[x]) <<" "<< imag(a[x]) << endl;
  f.close();
}

void Save(const char *fname, int n, double *cx, comple *a, comple *t) {
  ofstream f;
  f.open(fname);
  for(int x=0;x<n;x++) f << cx[x] <<" "<< norm(a[x]*t[x]) <<" "<< real(a[x]*t[x]) <<" "<< imag(a[x]*t[x]) << endl;
  f.close();
}


void SaveEField(const char *fname, int n, double *cx, comple *a, int b1, int b2, double e1, double e2) {
  int x;
  double ec;
  ofstream f;
  f.open(fname);

  ec = e2;
  for(x=0;x<=b1;x++) {
    f << cx[x] <<" "<< norm(a[x])/(ec*ec) <<" "<< real(a[x])/ec <<" "<< imag(a[x])/ec << endl;
  }

  ec = e1;
  for(x=b1;x<=b2;x++) {
    f << cx[x] <<" "<< norm(a[x])/(ec*ec) <<" "<< real(a[x])/ec <<" "<< imag(a[x])/ec << endl;
  }

  ec = e2;
  for(x=b2;x< n;x++) {
    f << cx[x] <<" "<< norm(a[x])/(ec*ec) <<" "<< real(a[x])/ec <<" "<< imag(a[x])/ec << endl;
  }

  f.close();
}

void Normalize(comple *a, int n, double scl) {
  int i;
  double nrm;

  nrm= 0.0;
  for(i=0;i<n;i++) if(nrm < norm(a[i])) nrm = norm(a[i]);
  if(nrm > 0.0) {
    nrm = 1.0/sqrt(nrm);
    for(i=0;i<n;i++) a[i] *= nrm*scl;
  }
  cout <<"Norm: " << nrm << endl;
}





complex<double> gauss1D(double wr, double x, double z, double k, double f) 
{
  complex<double> aux1, aux2, imagi(0.0,1.0);

  if( wr <= 0.0 ) return(1.0);

  if(f != 0.0) 
    aux1 =  1.0/(wr*wr) + imagi*k/(2.0*f);
  else
    aux1 =  1.0/(wr*wr);

  aux2 =  1.0/(1.0 + 2.0*imagi*z/k*aux1);
  return( sqrt(aux2)*exp(-x*x*aux1*aux2));
}

