#ifndef __cn_bpm_1D_h__
#define __cn_bpm_1D_h__

/*
Crank-Nicolson BPM solver for one-dimensional linear domain
*/

#include<iostream>
#include<fstream>
#include<cstdlib>
#include<complex>

#include"bpm_utils.h"

using namespace std;
typedef complex<double> comple;


class CNPropagator {

private:
public:

  // dimension
  int     N;

  // LP matrix, RHS, and auxiliary holders
  comple *A;         // lower sub diagonal
  comple *B;         // diagonal
  comple *C;         // upper diagonal
  comple *R;         // right-hand-side of the linear system to solve
  comple *X;         // used to be: permanent storage for diagonal

  comple *LL;        // discrete RHS operator coefficient arrays
  comple *LC;
  comple *LR;


  // electric field amplitude holders: we need two of these
  comple *E1;
  comple *E2;

  // coordinate array
  double *cx;

  // step auxiliaries
  comple  idelta;
  comple *currentE;    // pointer to where current solution resides (either E1 or E2)
  comple *auxiliaryE;  // this is where the new solution will be stored (either E1 or E2)

  // boundary conditions
  string bctype;

  // homogeneous constraints locations
  int    homcnd[16];

  // index map
  int BX1;
  int BX2;
  double eclad;
  double ecore;
  double refri;
  double gridx;


  CNPropagator() {
    A=B=C=R=E1=E2=X=0;
    homcnd[0] = 0;
  }

  void ReleaseMemory() {
    if(X !=0) delete[] X;
  }

 ~CNPropagator() {
   ReleaseMemory();
  }
  
  void Allocate(int n) {
    ReleaseMemory();

    N  = n;
    X  = new comple[N*11];
    cx = new double[N];

    if((X==0)||(cx==0)) {
      cerr <<"CNPropagator: failed to allocate\n";
      exit(1);
    }


    A  = &X[N*1];  
    B  = &X[N*2]; 
    C  = &X[N*3]; 
    R  = &X[N*4];  
    E1 = &X[N*5];  
    E2 = &X[N*6];  

    LL = &X[N*7];
    LC = &X[N*8];
    LR = &X[N*9];


    currentE   = E1;  
    auxiliaryE = E2;  
  }

  void SetDomain(double wx, double wc, double ncore, double nclad, bool verb=true) {

  eclad = nclad*nclad;
  ecore = ncore*ncore;
  refri = ncore;

  int nc = ((int) ceil(N*wc/wx));

  gridx = wc/((double) (nc-1));

  BX1 = (N - nc)/2;
  BX2 = BX1 + nc - 1;

  for(int x=0;x<N;x++) cx[x] = gridx*(x-N/2+0.5);

  if(verb) {
    cout << "domain:   " << wx    << endl;
    cout << "core:     " << wc    << endl;
    cout << "grid pts: " << N     << endl;
    cout << "core pts: " << nc    << endl;
    cout << "grid stp: " << gridx << endl;
    cout << "BX1:      " << BX1   << endl;
    cout << "BX2:      " << BX2   << endl;
    cout << "check:    " << (BX2-BX1)*gridx << endl;
  }
  }
  
  void Step(comple dz, double lambda) {
    int    row;
    comple scldz;
    double betar = 2.0*M_PI/lambda*refri;

    comple LBNDRY;
    comple RBNDRY;

    // half-step, as required in classic C-N:
    scldz = comple(0.0,1.0)*dz/(4.0*betar);

    // define bands in L_minus
    for(row=0;row<N;row++) {
      A[row] =      -scldz*LL[row];
      B[row] =  1.0 -scldz*LC[row];
      C[row] =      -scldz*LR[row];
    }
    
    // evaluate right-hand-side: L_plus E_old
    for(row=1;row<N-1;row++) {
      R[row]  = currentE[row];
      R[row] += scldz*LL[row]*currentE[row-1];
      R[row] += scldz*LC[row]*currentE[row+0];
      R[row] += scldz*LR[row]*currentE[row+1];
    }

    // PEC boundary conditions
    LBNDRY = 0.0;
    RBNDRY = 0.0;

    if( bctype == "ABC1" ) {
      LBNDRY = currentE[  0]/currentE[  1]*currentE[  0];
      RBNDRY = currentE[N-1]/currentE[N-2]*currentE[N-1];

      B[N-1] = B[N-1] + C[N-1]*currentE[N-1]/currentE[N-2];
      B[  0] = B[  0] + A[  0]*currentE[  0]/currentE[  1];
    }

    row = 0;
    R[row]  = currentE[row];
    R[row] += scldz*LL[row]*LBNDRY;
    R[row] += scldz*LC[row]*currentE[row+0];
    R[row] += scldz*LR[row]*currentE[row+1];
    
    row = N-1;
    R[row]  = currentE[row];
    R[row] += scldz*LL[row]*currentE[row-1];
    R[row] += scldz*LC[row]*currentE[row+0];
    R[row] += scldz*LR[row]*RBNDRY;

    // for conditions, zero RHS:
    for(int i=1;i<=homcnd[0];i++) {
      row = homcnd[i];
      A[row] = LL[row];
      B[row] = LC[row];
      C[row] = LR[row];
      R[row] = 0.0;
    }


    // Solve the linear system defined by L_minus:
    TDMsolver(N, A, B, C, R, auxiliaryE);

    // 3. update what is current amplitude holder
    comple *aux = currentE;
    currentE   = auxiliaryE;
    auxiliaryE = aux;
  }

  comple* GetAmplitudePtr() {
    return( currentE );
  }


  double* GetCoordinatesX() {
    return( cx );
  }

  void SetScheme(double lambda, double refindex, string scheme, void *par = 0) {
    refri = refindex;

    if(scheme=="TM") {
      SetRHS_TM(LL, LC, LR, N, BX1, BX2, gridx, lambda, refindex, eclad, ecore);
      return;
    }

    if(scheme=="TE") {
      SetRHS_TE(LL, LC, LR, N, BX1, BX2, gridx, lambda, refindex, eclad, ecore);
      return;
    }

    cout <<"Stop: Discretization scheme name not recognized" << endl;
    cout <<"Expected: [TM,TE]" << endl;
    exit(1);
  }

};


#endif // __cn_bpm_1D_h__
