/*
Crank-Nicolson BPM illustrated - one-dimensional linear domain
*/

#include<iostream>
#include<fstream>
#include<cstdlib>
#include<complex>

using namespace std;
typedef complex<double> comple;

#include"parameter_input_output.h"
#include"bpm_utils.h"
#include"cn-bpm-1D.h"


class Params : public NamedParameterList {
public:

  namint nx;
  namdbl wx;
  namdbl wc;
  namdbl w0;
  namdbl la;
  namdbl zi;
  namdbl zr;
  namdbl dz;
  namdbl ni;
  namdbl no;
  namdbl nr;

  namstr pl;

  Params() {
    SetName("TestParameters");
    IncludeParameter(nx,"NX");
    IncludeParameter(wx,"LX");
    IncludeParameter(wc,"Lcore");
    IncludeParameter(ni,"index_core");
    IncludeParameter(no,"index_clad");
    IncludeParameter(nr,"index_reff");
    IncludeParameter(la,"lambda");
    IncludeParameter(zi,"distance_imag");
    IncludeParameter(zr,"distance_real");
    IncludeParameter(dz,"z_step");
    IncludeParameter(w0,"initial_w0");
    IncludeParameter(pl,"polarization");
  }
};




int main(int argc, char *argv[]) {
  int x;

  if(argc < 2) {
    cout <<"input file-name expected" << endl;
    exit(1);
  }

  Params pars;
  pars.Read(argv[1]);
  pars.Write(cout);

  // computational domain params
  double wx    = pars.wx.v();
  int    nx    = pars.nx.v();

  // waveguide params
  double wc    = pars.wc.v();
  double ncore = pars.ni.v();
  double nclad = pars.no.v();
  double nreff = pars.nr.v();

  // initial condition params
  double la = pars.la.v();
  double w0 = pars.w0.v();

  // driver params
  double dz       = pars.dz.v();
  double fz_imag  = pars.zi.v();
  double fz_real  = pars.zr.v();
  double zcurrent = 0.0;

  // my solver
  CNPropagator cnp;

  // allocate memory
  cnp.Allocate(nx);
  cnp.SetDomain(wx, wc, ncore, nclad);
  cnp.SetScheme(la,nreff,pars.pl.v());

  double dx = cnp.gridx;

  // set initial condition
  for(x=0;x<nx;x++) cnp.GetAmplitudePtr()[x] = gauss1D(w0, dx*(x - nx/2), 0.0, 2.0*M_PI/la, 0.0);
  SaveEField("sol-i.dat",nx,cnp.GetCoordinatesX(),cnp.GetAmplitudePtr(), cnp.BX1, cnp.BX2, cnp.ecore, cnp.eclad);
  
  // evolution in imaginary distance
  for(zcurrent=0.0;zcurrent<fz_imag;) {
    cnp.Step(comple(0.0,-dz), la);
    zcurrent += dz;
  }

  // Note: the first option assumes D in array, normalizes for E 
  if(pars.pl.v() == "TM")
    Normalize(cnp.GetAmplitudePtr(),nx, ncore*ncore);
  else
    Normalize(cnp.GetAmplitudePtr(),nx, 1.0);

  // show modal solution: 
  if(pars.pl.v() == "TM")  
    SaveEField("sol-m.dat",nx,cnp.GetCoordinatesX(),cnp.GetAmplitudePtr(), cnp.BX1, cnp.BX2, cnp.ecore, cnp.eclad);

  if(pars.pl.v() == "TE") 
    Save("sol-m.dat",nx,cnp.GetCoordinatesX(),cnp.GetAmplitudePtr());


  comple aold;
  comple anew;
  comple prop;
  double neff;
  int cnt = 0;

  aold = cnp.GetAmplitudePtr()[nx/2];

  // evolution in real distance
  for(zcurrent=0.0;zcurrent<fz_real;) {
    cnp.Step(comple(dz,0.0), la);
    zcurrent += dz;
    cnt++;

    if(cnt==20) {
      anew = cnp.GetAmplitudePtr()[nx/2];
      prop = log(anew/aold)/(cnt*dz);
      neff = nreff + imag(prop)*la/(2.0*M_PI);
      
      cout << zcurrent <<" "<< prop <<" "<< norm(anew/aold)-1.0 <<" "<< neff << endl;

      aold = anew;
      cnt = 0;
    }

  }

  // final numerical solution
  if(pars.pl.v() == "TM") 
    SaveEField("sol-f.dat",nx,cnp.GetCoordinatesX(),cnp.GetAmplitudePtr(), cnp.BX1, cnp.BX2, cnp.ecore, cnp.eclad);

  if(pars.pl.v() == "TE") 
    Save("sol-f.dat",nx,cnp.GetCoordinatesX(),cnp.GetAmplitudePtr());
}



